package net.shrine.messagequeuemiddleware

import cats.effect.IO
import net.shrine.config.ConfigSource
import net.shrine.hub.data.store.HubDb
import net.shrine.protocol.i2b2.ResultOutputType
import net.shrine.protocol.i2b2.query.{I2b2QueryDefinition, ValueConstraint}
import net.shrine.protocol.version.v2.querydefinition.{Concept, ConceptConstraint, ConceptGroup, FlagConstraints, QueryDefinition}
import net.shrine.protocol.version.{DateStamp, Envelope, ItemVersion, JsonText, MomQueueName, NodeKey, ProtocolVersion, QueryId, ResultId, ShrineVersion}
import net.shrine.protocol.version.v2.{Breakdowns, Node, ObfuscatingParameters, Query, QueryProgress, QueryStatus, Researcher, Result, ResultProgress, ResultStatus, RunQueryAtHub, RunQueryForResult, UpdateQueryAtQep, UpdateQueryAtQepWithStatus, UpdateQueryReadyForAdapters, UpdateResult, UpdateResultWithCount, UpdateResultWithError, UpdateResultWithProgress, VersionInfo}
import net.shrine.protocol.version.v1.{
  Breakdowns => BreakdownsV1,
  CrcResult => CrcResultV1,
  ErrorResult => ErrorResultV1,
  Node => NodeV1,
  Query => QueryV1,
  QueryStatuses => QueryStatusesV1,
  Researcher => ResearcherV1,
  Result => ResultV1,
  ResultOutputType => ResultOutputTypeV1,
  ResultProgress => ResultProgressV1,
  ResultStatus => ResultStatusV1,
  ResultStatuses => ResultStatusesV1,
  RunQueryAtHub => RunQueryAtHubV1,
  RunQueryForResult => RunQueryForResultV1,
  Topic => TopicV1,
  TopicId => TopicIdV1,
  UpdateQueryAtQep => UpdateQueryAtQepV1,
  UpdateQueryAtQepWithStatus => UpdateQueryAtQepWithStatusV1,
  UpdateQueryReadyForAdapters => UpdateQueryReadyForAdaptersV1,
  UpdateResult => UpdateResultV1,
  UpdateResultWithCrcResult => UpdateResultWithCrcResultV1,
  UpdateResultWithError => UpdateResultWithErrorV1,
  VersionInfo => VersionInfoV1
}

import scala.xml.NodeSeq

abstract class MessageTranslationStrategy {
  def translateMesssage(toQueueName: MomQueueName, envelope: Envelope): IO[(MomQueueName, Envelope)]
}

class RunQueryForResultStrategyV1 extends MessageTranslationStrategy {
  def translateMesssage(toQueueName: MomQueueName, envelope: Envelope): IO[(MomQueueName, Envelope)] = {

    val runQueryForResultV1: RunQueryForResultV1 = RunQueryForResultV1.tryRead (new JsonText (envelope.contents) ).get
    HubDb.db.selectNodeIO(runQueryForResultV1.node.id).flatMap (nodeOpt => {
      val destNode = nodeOpt.head

      if (destNode.protocolVersion.underlying == 2) {

        //val resultOutputTypes: Set[ResultOutputType] = ResultOutputType.fromBreakdownNames(d.query.breakdownNames)

        val query: Query = Query.create(
          id= runQueryForResultV1.query.id.underlying,
          queryDefinition = TransformV1QueryToV2Query.process(I2b2QueryDefinition.fromI2b2(runQueryForResultV1.query.queryDefinitionXml).get),
          breakdownNames = Seq(),
          queryName = runQueryForResultV1.query.queryName,
          nodeOfOriginId = runQueryForResultV1.query.nodeOfOriginId,
          researcherId = runQueryForResultV1.researcher.id,
        )

        val node: Node = new Node(
          id= runQueryForResultV1.node.id,
          name = runQueryForResultV1.node.name,
          key = runQueryForResultV1.node.key,
          momQueueName = new MomQueueName(runQueryForResultV1.node.key.underlying),
          userDomainName = runQueryForResultV1.node.userDomainName,
          adminEmail = runQueryForResultV1.node.adminEmail,
          momId = destNode.momId
        )

        val researcher: Researcher = new Researcher(
          id = runQueryForResultV1.researcher.id,
          userName = runQueryForResultV1.researcher.userName,
          userDomainName = runQueryForResultV1.researcher.userDomainName,
          nodeId = node.id
        )

        val resultProgress: ResultProgress = Result.create(query, node)

        val runQueryForResult = RunQueryForResult(query, researcher, node, resultProgress, new ProtocolVersion(1))

        val resultEnvelope: Envelope = Envelope(RunQueryForResult.envelopeType, runQueryForResultV1.query.id.underlying, runQueryForResult.asJsonText.underlying,new ProtocolVersion(1), None)

        IO((destNode.momQueueName, resultEnvelope))
      } else {
        IO((toQueueName, envelope))
      }
    })
  }
}

class RunQueryForResultStrategyV2 extends MessageTranslationStrategy {
  def translateMesssage(toQueueName: MomQueueName, envelope: Envelope): IO[(MomQueueName, Envelope)] = {

    val d: RunQueryForResult = RunQueryForResult.tryRead (new JsonText (envelope.contents) ).get
    HubDb.db.selectNodeIO (d.node.id).flatMap (nodeOpt => {
      val destNode = nodeOpt.get

      if (destNode.protocolVersion.underlying == 1) {

        def countAndFilteredBreakdowns(selectedBreakDowns: Set[ResultOutputType]): NodeSeq = {
          val filteredResultOutputTypes = Seq(ResultOutputType.PATIENT_COUNT_XML) ++ selectedBreakDowns.toSeq.sortBy(_.name)

          <result_output_list>
            {filteredResultOutputTypes.zipWithIndex.map { typeAndIndex => {
              <result_output priority_index={(typeAndIndex._2 + 1).toString} name={typeAndIndex._1.name.toLowerCase}/>
          }
          }}
          </result_output_list>
        }

        val resultOutputTypes: Set[ResultOutputType] = ResultOutputType.fromBreakdownNames(d.query.breakdownNames)

        val queryV1: QueryV1 = QueryV1.create(
          id= d.query.id.underlying,
          queryDefinitionXml = I2b2QueryDefinition.fromShrineV2(d.query.queryDefinition, d.query.id, d.query.queryName).toI2b2.toString(),
          outputTypesXml = countAndFilteredBreakdowns(resultOutputTypes).toString(),
          queryName = d.query.queryName,
          topicId = new TopicIdV1(1L),
          nodeOfOriginId = d.query.nodeOfOriginId,
          researcherId = d.researcher.id,
          projectName = "SHRINE",
          versionInfo = VersionInfoV1(
            protocolVersion = new ProtocolVersion(1),
            itemVersion = d.query.versionInfo.itemVersion,
            createDate = DateStamp.now,
            changeDate = DateStamp.now
          )
        )

        val nodeV1: NodeV1 = NodeV1(
          id= d.node.id,
          versionInfo = VersionInfoV1(
          protocolVersion = new ProtocolVersion(1),
          itemVersion = d.node.versionInfo.itemVersion,
          createDate = DateStamp.now,
          changeDate = DateStamp.now
        ),
          name = d.node.name,
          key = d.node.key,
          momQueueName = new MomQueueName(d.node.key.underlying),
          userDomainName = d.node.userDomainName,
          adminEmail = d.node.adminEmail
        )

        val researcherV1: ResearcherV1 = ResearcherV1(
          id= d.researcher.id,
          versionInfo = VersionInfoV1(
            protocolVersion = new ProtocolVersion(1),
            itemVersion = d.researcher.versionInfo.itemVersion,
            createDate = DateStamp.now,
            changeDate = DateStamp.now
          ),
          userName = d.researcher.userName,
          userDomainName = d.researcher.userDomainName,
          nodeId = d.node.id
        )
        val resultProgressV1: ResultProgressV1 = ResultV1.create(queryV1, nodeV1)
        val topicV1: TopicV1 = TopicV1.createCompatibleWithShrine200(
          researcherId = d.researcher.id,
          name="DefaultName",
          description="Default Description",
          localId = queryV1.topicId.underlying.toInt
        )

        val runQueryForResultV1 = RunQueryForResultV1(queryV1, researcherV1, nodeV1, topicV1, resultProgressV1, new ProtocolVersion(1))

        val resultEnvelope: Envelope = Envelope(RunQueryForResult.envelopeType, d.query.id.underlying, runQueryForResultV1.asJsonText.underlying,new ProtocolVersion(1), None)

        IO((destNode.momQueueName, resultEnvelope))
      } else {
        IO((toQueueName, envelope))
      }
    })
  }
}

class UpdateResultStrategyV1 extends MessageTranslationStrategy {
  def translateMesssage(toQueueName: MomQueueName, envelope: Envelope): IO[(MomQueueName, Envelope)] = {
    envelope match {
      case envelope: Envelope if envelope.protocolVersion.underlying != 1 => IO((toQueueName, envelope))
      case _ =>
        val updateResult: UpdateResultV1 = UpdateResultV1.tryRead(new JsonText(envelope.contents)).get

        HubDb.db.selectNodeByKeyIO(updateResult.adapterNodeKey).flatMap(nodeTry => {
          val adapterNode: Node = nodeTry.get
          val status = ResultStatus.namesToStatuses(updateResult.status.statusName)
          val queryId = new QueryId(updateResult.queryId.underlying)
          //get the result Id
          lookupResultId(queryId, adapterNode).flatMap(resultId => {
            val newUpdatedResult = ResultProgress(
              id = resultId,
              queryId = queryId,
              adapterNodeId = adapterNode.id,
              adapterNodeName = adapterNode.name
            )
            if (status.isFinal && !status.isError) {
              val updateResultWithCrcResult: UpdateResultWithCrcResultV1 = updateResult.asInstanceOf[UpdateResultWithCrcResultV1]

              val obfuscatingParams = ObfuscatingParameters.fromConfig(ConfigSource.config.getConfig("shrine.hub.messagequeue.obfuscation"))

              val breakdowns = updateResultWithCrcResult.breakdowns.map(breakdowns => {
                Breakdowns( breakdowns.counts)
              })
              val newUpdatedResultWithCount = newUpdatedResult.toCrcResult(
                count = updateResultWithCrcResult.count,
                crcQueryInstanceId = updateResultWithCrcResult.crcQueryInstanceId,
                statusMessage = updateResultWithCrcResult.statusMessage,
                breakdowns = breakdowns,
                obfuscatingParameters = obfuscatingParams,
              )
              val resultProgress = UpdateResult.createUpdateResult(newUpdatedResultWithCount)
              val resultEnvelope: Envelope = Envelope(UpdateResult.envelopeType, updateResult.queryId.underlying, resultProgress.asJsonText.underlying)

              IO((toQueueName, resultEnvelope))
            }
            else if(status.isError){
              val updateResultWithError: UpdateResultWithErrorV1 = updateResult.asInstanceOf[UpdateResultWithErrorV1]

              val newUpdatedResultWithError = newUpdatedResult.toErrorFromJsonProblemDigest(
                problem = updateResultWithError.problem,
                status =  ResultStatus.namesToStatuses(updateResultWithError.status.statusName),
                statusMessage = updateResultWithError.statusMessage,
                crcQueryInstanceId = updateResultWithError.crcQueryInstanceId,
                adapterTime= DateStamp.now
              )
              val resultError = UpdateResult.createUpdateResult(newUpdatedResultWithError)
              val resultEnvelope: Envelope = Envelope(UpdateResult.envelopeType, updateResult.queryId.underlying, resultError.asJsonText.underlying)

              IO((toQueueName, resultEnvelope))
            }
            else {
              val newUpdatedResultWithStatus = newUpdatedResult.withStatus(status)
              val resultProgress = UpdateResult.createUpdateResult(newUpdatedResultWithStatus)
              val resultEnvelope: Envelope = Envelope(UpdateResult.envelopeType, updateResult.queryId.underlying, resultProgress.asJsonText.underlying)

              IO((toQueueName, resultEnvelope))
            }
          })
        })
    }
  }
  def lookupResultId(queryId: QueryId, adapterNode: Node): IO[ResultId] ={
    HubDb.db.selectMostRecentResultsForQueryIO(queryId).map(results => {
      results.filter(_.adapterNodeId == adapterNode.id).head.id
    })
  }
}

class UpdateResultStrategyV2 extends MessageTranslationStrategy {
  def translateMesssage(toQueueName: MomQueueName, envelope: Envelope): IO[(MomQueueName, Envelope)] = {
    val updateResult: UpdateResult = UpdateResult.tryRead(new JsonText(envelope.contents)).get

    HubDb.db.selectLatestNodesIO.flatMap(nodes => {

      val filteredNodes = nodes.filter(_.get.momQueueName == toQueueName)

      if (filteredNodes.isEmpty) {
        IO((toQueueName, envelope))
      } else {
        val adapterNode = filteredNodes.head.get

        if (adapterNode.protocolVersion.underlying == 1) {
          val status: ResultStatusV1 =
            ResultStatusesV1.namesToStatuses(updateResult.result.status.statusName)
          val queryId = new QueryId(updateResult.result.queryId.underlying)

          lookupResultId(queryId, adapterNode).flatMap(resultId => {
            val version1Info: VersionInfoV1 = VersionInfoV1(
              protocolVersion = new ProtocolVersion(1),
              itemVersion = updateResult.result.versionInfo.itemVersion,
              createDate = DateStamp.now,
              changeDate = DateStamp.now
            )
            if (status.isFinal && !status.isError) {
              val updateResultWithCount: UpdateResultWithCount =
                updateResult.asInstanceOf[UpdateResultWithCount]

              val breakdowns: Option[BreakdownsV1] = updateResultWithCount.result.breakdowns.map(breakdowns => {
                BreakdownsV1(breakdowns.counts)
              })

              val resultProgress2: CrcResultV1 =  CrcResultV1(
                  id = resultId,
                  versionInfo = version1Info,
                  queryId = updateResultWithCount.result.queryId,
                  adapterNodeId = updateResultWithCount.result.adapterNodeId,
                  adapterNodeName = updateResultWithCount.result.adapterNodeName,
                  count = updateResultWithCount.result.count,
                  crcQueryInstanceId = updateResultWithCount.result.crcQueryInstanceId,
                  resultType = ResultOutputTypeV1.PATIENT_COUNT_XML,
                  breakdowns = breakdowns,
                  status = ResultStatusesV1.namesToStatuses(updateResultWithCount.result.status.statusName),
                  statusMessage = None
                )
              val resultEnvelope: Envelope = Envelope(ResultV1.envelopeType, updateResult.result.queryId.underlying, resultProgress2.asJsonText.underlying, new ProtocolVersion(1))

              IO((toQueueName, resultEnvelope))
            }
            else if (status.isError) {
              val updateResultWithError: UpdateResultWithError = updateResult.asInstanceOf[UpdateResultWithError]

              val resultError = ErrorResultV1(
                id = resultId,
                versionInfo= version1Info,
                queryId = updateResultWithError.result.queryId,
                adapterNodeId = updateResultWithError.result.adapterNodeId,
                adapterNodeName  = updateResultWithError.result.adapterNodeName,
                status = ResultStatusesV1.namesToStatuses(updateResultWithError.result.status.statusName),
                statusMessage = updateResultWithError.result.statusMessage,
                problemDigest = updateResultWithError.result.problemDigest
              )

              val resultEnvelope: Envelope = Envelope(ResultV1.envelopeType,
                resultError.queryId.underlying,
                resultError.asJsonText.underlying,
                new ProtocolVersion(1))

              IO((toQueueName, resultEnvelope))
            }
            else {
              val updateResultWithProgress: UpdateResultWithProgress = updateResult.asInstanceOf[UpdateResultWithProgress]

              val resultProgress = ResultProgressV1(
                id = resultId,
                versionInfo = version1Info,
                queryId = updateResultWithProgress.result.queryId,
                adapterNodeId = updateResultWithProgress.result.adapterNodeId,
                adapterNodeName= updateResultWithProgress.result.adapterNodeName,
                status = ResultStatusesV1.namesToStatuses(updateResultWithProgress.result.status.statusName),
                statusMessage = updateResultWithProgress.result.statusMessage,
                crcQueryInstanceId = updateResultWithProgress.result.crcQueryInstanceId
              )

              val resultEnvelope: Envelope = Envelope(ResultV1.envelopeType,
                updateResultWithProgress.result.queryId.underlying, resultProgress.asJsonText.underlying, new ProtocolVersion(1))
              IO((toQueueName, resultEnvelope))
            }
          })
        }
        else{
          IO((toQueueName, envelope))
        }
      }
    })
  }

  def lookupResultId(queryId: QueryId, adapterNode: Node): IO[ResultId] ={
    HubDb.db.selectMostRecentResultsForQueryIO(queryId).map(results => {
      results.filter(_.adapterNodeId == adapterNode.id).head.id
    })
  }
}


class RunQueryAtHubStrategyV1 extends MessageTranslationStrategy {
  def translateMesssage(toQueueName: MomQueueName, envelope: Envelope): IO[(MomQueueName, Envelope)] = {
  val d: RunQueryAtHubV1 = RunQueryAtHubV1.tryRead (new JsonText (envelope.contents) ).get
    HubDb.db.selectNodeByKeyIO(NodeKey.localNodeKey).flatMap (nodeOpt => {
      val adapterNode = nodeOpt.head
      val version2Info = VersionInfo(
        protocolVersion = ProtocolVersion.current,
        shrineVersion = ShrineVersion.current,
        itemVersion = ItemVersion.one,
        createDate = DateStamp.now,
        changeDate = DateStamp.now
      )
      val query: Query = Query.create(
        id= d.query.id.underlying,
        queryDefinition = TransformV1QueryToV2Query.process(I2b2QueryDefinition.fromI2b2(d.query.queryDefinitionXml).get),
        breakdownNames = Seq(),
        queryName = d.query.queryName,
        nodeOfOriginId = d.query.nodeOfOriginId,
        researcherId = d.researcher.id,
      )
      val researcher: Researcher = new Researcher(
        id = d.researcher.id,
        userName = d.researcher.userName,
        userDomainName = d.researcher.userDomainName,
        nodeId = d.researcher.nodeId
      )
      val queryProgress: QueryProgress = QueryProgress(
        id = query.id,
        versionInfo= version2Info,
        status = QueryStatus.namesToStatuses(d.query.status.statusName),
        queryDefinition= query.queryDefinition,
        breakdownNames = query.breakdownNames,
        queryName = query.queryName,
        queryNotes = None,
        queryFaved= false,
        nodeOfOriginId = query.nodeOfOriginId,
        researcherId = query.researcherId
      )

      val runQueryAtHub: RunQueryAtHub = RunQueryAtHub(
        queryProgress,
        researcher
      )
      val resultEnvelope: Envelope = Envelope(RunQueryAtHub.envelopeType, d.query.id.underlying, runQueryAtHub.asJsonText.underlying, ProtocolVersion.current)

      IO((toQueueName, resultEnvelope))
    })
  }
}

class UpdateQueryAtQepStrategyV2 extends MessageTranslationStrategy {
  def translateMesssage(toQueueName: MomQueueName, envelope: Envelope): IO[(MomQueueName, Envelope)] = {
    val updateQueryAtQep: UpdateQueryAtQep = UpdateQueryAtQep.tryRead (new JsonText (envelope.contents) ).get
    HubDb.db.selectLatestNodesIO.map(nodes => nodes.filter(p => p.get.momQueueName==toQueueName)).flatMap (nodeOpt => {
      val node = nodeOpt.head.get
      if(node.protocolVersion.underlying == 1) {
        updateQueryAtQep match {
          case statusUpdate: UpdateQueryAtQepWithStatus =>
            val updateQueryAtQepWithStatus: UpdateQueryAtQepWithStatusV1 =
              UpdateQueryAtQepWithStatusV1(
              statusUpdate.queryId,
              QueryStatusesV1.namesToStatuses(statusUpdate.queryStatus.statusName),
              changeDate = DateStamp.now
            )

            val resultEnvelope: Envelope = Envelope(UpdateQueryAtQepV1.envelopeType,
              statusUpdate.queryId.underlying,
              updateQueryAtQepWithStatus.asJsonText.underlying,
              new ProtocolVersion(1)
            )

            IO((toQueueName, resultEnvelope))

          case statusUpdate: UpdateQueryReadyForAdapters =>
            val updateQueryReadyForAdapters: UpdateQueryReadyForAdaptersV1 = UpdateQueryReadyForAdaptersV1(
            queryId = statusUpdate.queryId,
            changeDate= DateStamp.now,
            resultProgresses = statusUpdate.resultProgresses.map(p => {
             ResultProgressV1(
                id = p.id,
                versionInfo = VersionInfoV1(
                  protocolVersion = new ProtocolVersion(1),
                  itemVersion = p.versionInfo.itemVersion,
                  createDate = DateStamp.now,
                  changeDate = DateStamp.now
                ),
                queryId = statusUpdate.queryId,
                adapterNodeId = p.adapterNodeId,
                adapterNodeName = p.adapterNodeName,
                status = ResultStatusesV1.namesToStatuses(p.status.statusName),
                statusMessage = p.statusMessage,
                crcQueryInstanceId= p.crcQueryInstanceId
              )})
            )

            val resultEnvelope: Envelope = Envelope(UpdateQueryAtQepV1.envelopeType,
              statusUpdate.queryId.underlying,
              updateQueryReadyForAdapters.asJsonText.underlying,
              new ProtocolVersion(1)
            )

            IO((toQueueName, resultEnvelope))

          case _  => IO((toQueueName, envelope))
        }
      }
      else{
        IO((toQueueName, envelope))
      }
    })
  }
}

object MessageTranslationStrategy {
  val v1 = new ProtocolVersion(1)
  val v2 = new ProtocolVersion(2)

  val strategies = Map(
    (v2,"RunQueryForResult") -> new RunQueryForResultStrategyV2,
    (v1,"RunQueryAtHub") -> new  RunQueryAtHubStrategyV1,
    (v1,"UpdateResult") -> new UpdateResultStrategyV1,
    (v2,"UpdateResult") -> new UpdateResultStrategyV2,
    (v2,"UpdateQueryAtQep") -> new  UpdateQueryAtQepStrategyV2,
  )
}

object TransformV1QueryToV2Query {
  def process(i2b2QueryDefinition: I2b2QueryDefinition): QueryDefinition = {
    val conceptGroups = i2b2QueryDefinition.expr.map(ex => {
      I2b2QueryDefinition.toPanels(ex).map(panel => {
        val startDate = panel.start.map(date => new DateStamp(date.toGregorianCalendar.getTime.getTime))
        val endDate = panel.end.map(date => new DateStamp(date.toGregorianCalendar.getTime.getTime))
        val conceptsWithOutConstraints = panel.terms.filter(_.isTerm).map(texpr => {
          Concept(texpr.name, texpr.value)
        })
        val conceptsWithConstraints = panel.termsWithConstraints.filter(_.isTerm).map(texpr => {
          val conceptConstraint: Option[ConceptConstraint] =  texpr.valueConstraint.map {
              case vCon: ValueConstraint if vCon.valueType == "FLAG"  && vCon.value == "L" => FlagConstraints.Low
              case vCon: ValueConstraint if vCon.valueType == "FLAG"  && vCon.value == "@" => FlagConstraints.Normal
              case vCon: ValueConstraint if vCon.valueType == "FLAG"  && vCon.value == "H" => FlagConstraints.High
          }
          Concept(texpr.term.name, texpr.term.value, conceptConstraint)
        })

        val allConcepts = conceptsWithOutConstraints ++ conceptsWithConstraints
        val conceptGroup = if (panel.isExcluded) ConceptGroup.noneOf(allConcepts) else ConceptGroup.atLeastOneOf(allConcepts)

        val groupWithStartDate = startDate.fold(conceptGroup){date => conceptGroup.withStartDate(date)}
        val groupWithEndDate =   endDate.fold(groupWithStartDate){date => groupWithStartDate.withEndDate(date)}
        val groupWithOccurs = if(panel.minOccurrences > 1) groupWithEndDate.withOccursAtLeast(panel.minOccurrences) else groupWithEndDate

        groupWithOccurs
      })
    }).get
    QueryDefinition.allOf(conceptGroups)
  }
}
